/**
 ****************************************************************************************
 *
 * @file usb_smsd.c
 *
 * @brief USB SmartMSD app implementation
 *
 * Copyright (c) 2018 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */
#include "ad_nvms.h"
#include "sys_charger.h"
#include "sys_power_mgr.h"
#include "sys_watchdog.h"
#include "hw_usb.h"
#include "USB.h"
#include "USB_MSD.h"
#include "USB_SmartMSD.h"

/*********************************************************************
 *
 *       Defines, configurable
 *
 **********************************************************************
 */
#define SMSD_USE_NVMS //Use NVMS or RAM for data
#define SMSD_DATA_SIZE 64*5*1024 //file size

/**
 * Define the number of sectors (512bytes) in the mounted drive
 * FW_UPDATE partition is from 0x90000 to 0xE1000 (ie 0x51000 bytes)
 * FW_EXEC partition is from 0x20000 to 0x70000 (ie 0x50000 bytes)
 * So take the smaller of these as the size of the partition
 * Add 32 sectors (16k) for the FAT table
 */
#define _SMARTMSD_NUM_SECTORS (32+(64*5)*2)
// Only 1 file UPDATE.DAT defined
#define MAX_CONST_FILE 1
//#define APP_FILE_HEADER "FWBIN"
#define usb_main_TASK_PRIORITY              ( OS_TASK_PRIORITY_NORMAL )

__RETAINED static OS_TASK usb_smsd_task_handle;
__RETAINED static uint8 run_usb_task/* = 0*/;
static uint8 _ReattchRequest;
static USB_HOOK UsbpHook;
static bool isFwFile;

/**
 *  Information that is used during enumeration.
 */
static const USB_DEVICE_INFO _DeviceInfo = {
        0x2DCF,         // VendorId
        0x6004,         // ProductId
        "Dialog Semiconductor",       // VendorName
        "DA1468x/DA1510x SmartMSD device",  // ProductName
        "0123456789AB" // SerialNumber. Should be 12 character or more for compliance with Mass Storage Device Bootability spec.
        };
/**
 * String information used when inquiring the volume 0.
 */
static const USB_MSD_LUN_INFO _Lun0Info = {
        "Vendor",     // MSD VendorName
        "MSD Volume", // MSD ProductName
        "1.00",       // MSD ProductVer
        "134657890"   // MSD SerialNo
        };


/**
 * @brief Application idle task hook
 */
uint8 isUSBActive(void)
{
  return run_usb_task;
}


/**
 * @brief Application idle task hook
 */
#ifdef SMSD_USE_NVMS
static int usb_read_from_nvms(nvms_partition_id_t nv_id, uint8* pData, uint32 Off, uint32 Numbytes)
{
        nvms_t nvms;

        nvms = ad_nvms_open(nv_id);
        if (nvms) {
                ad_nvms_read(nvms, Off, pData, Numbytes);
                return 0;
        }
        return -1;
}

/**
 * @brief Application idle task hook
 */
static int usb_write_to_nvms(nvms_partition_id_t nv_id, uint8* pData, uint32 Off, uint32 Numbytes)
{
        nvms_t nvms;

        nvms = ad_nvms_open(nv_id);
        if (nvms) {
                ad_nvms_write(nvms, Off, pData, Numbytes);
                ad_nvms_flush(nvms, true);
                return 0;
        }
        return -1;
}
#endif


/**
 * @brief Write Callback
 */
static int _cbOnWrite(unsigned Lun, const U8* pData, U32 Off, U32 NumBytes,
        const USB_SMSD_FILE_INFO* pFile)
{
        int write_bytes;

        if (NumBytes == 0) {
                return 0;
        }
        if (Off == 0) {
                isFwFile = true;
        }

        if ((Off + NumBytes) > SMSD_DATA_SIZE) {
                write_bytes = SMSD_DATA_SIZE - Off;
        } else {
                write_bytes = NumBytes;
        }

        if ((strncmp("UPDATE", (char *)pFile->pDirEntry->ShortEntry.acFilename, 6) == 0)
                && (strncmp("DAT", (char *)pFile->pDirEntry->ShortEntry.acExt, 3) == 0)) {

                if ((Off + write_bytes) <= SMSD_DATA_SIZE) {
                        usb_write_to_nvms(NVMS_FW_UPDATE_PART, (uint8 *)pData, Off, write_bytes);
                }
                else {
                        _ReattchRequest = 1;
                        USBD_MSD_RequestRefresh(0, USB_MSD_RE_ATTACH | USB_MSD_TRY_DISCONNECT);
                        return -1;
                }
        }

        return 0;
}

/**
 * @brief Read Callback
 */
static int _cbOnRead(unsigned Lun, U8* pData, U32 Off, U32 NumBytes,
        const USB_SMSD_FILE_INFO* pFile)
{
        (void)Lun;
        (void)Off;
        (void)NumBytes;
        (void)pFile;

        if (NumBytes == 0 || pData == 0)
                return 0;
        if ((strncmp("UPDATE", (char *)pFile->pDirEntry->ShortEntry.acFilename, 6) == 0)
                && (strncmp("DAT", (char *)pFile->pDirEntry->ShortEntry.acExt, 3) == 0)) {
                usb_read_from_nvms(NVMS_FW_UPDATE_PART, (uint8 *)pData, Off, NumBytes);
        }

        return 0;
}

/**
 * @brief Wrapper for memory allocation
 */
static void *_memAlloc(U32 size)
{
        return OS_MALLOC(size);
}

/**
 * @brief Wrapper for memory de-allocation
 */
static void _memFree(void *p)
{
        OS_FREE(p);
}


/**
 * SMSD User API definition
 */
static const USB_SMSD_USER_FUNC_API _UserFuncAPI = {
        _cbOnRead,     // pfOnRead    -> Is called when a sector of a given file is read.
        _cbOnWrite,    // pfOnWrite   -> Is called when a sector of a given file is written.
        _memAlloc,     // pfMemAlloc  -> Optional, can be set in order to allow the SmartMSD to share the mem alloc function of a system.
        _memFree       // pfMemFree   -> Optional, can be set in order to allow the SmartMSD to share the mem free function of a system.
        };




/**
 * @brief called by the USB MSD Module during USB_SmartMSD_Init() and initializes the SmartMSD volume.
 */
void USB_SMSD_X_Config(void)
{
        //
        // Global configuration
        //
        USBD_SMSD_SetUserAPI(&_UserFuncAPI);

        //
        // Setup LUN0
        //
        USBD_SMSD_SetNumSectors(0, _SMARTMSD_NUM_SECTORS);
        USBD_SMSD_SetSectorsPerCluster(0, 4); // Anywhere from 1 ... 128, but needs to be a Power of 2
        USBD_SMSD_SetNumRootDirSectors(0, 2);
        USBD_SMSD_SetVolumeInfo(0, "Virt0.MSD", &_Lun0Info);   // Add volume ID

        // No constant files are in the volume, so no need to call USBD_SMSD_AddConstFiles()
}

/**
 * @brief tracking the USB state for debug purpose
 */
void usb_smsd_state_cb(void * pContext, U8 NewState)
{
        printf("Newstate %x ",NewState);
        if (NewState & USB_STAT_ATTACHED) {
                //Attached
                printf("attached\n\r");
        }

        if (NewState & USB_STAT_READY) {
                //Ready
                printf("ready\n\r");
        }

        if (NewState & USB_STAT_ADDRESSED) {
                //Addressed
                printf("addressed\n\r");
        }

        if (NewState & USB_STAT_CONFIGURED) {
                //Configured
                printf("configured\n\r");
        }

        if (NewState & USB_STAT_SUSPENDED) {
                //Suspended
                printf("suspended\n\r");
        }
}


/**
 * @brief USB Mass storage main task
 */
void usb_smsd_task(void *params)
{
        USBD_Init();
        USBD_SMSD_Init();
        USBD_RegisterSCHook(&UsbpHook, usb_smsd_state_cb, NULL);
        USBD_SMSD_Add();
        USBD_SetDeviceInfo(&_DeviceInfo);
        USBD_Start();

        while (1) {
                //
                // Wait for configuration
                //
                while ((USBD_GetState() & (USB_STAT_CONFIGURED | USB_STAT_SUSPENDED))
                        != USB_STAT_CONFIGURED) {
                        USB_OS_Delay(50);
                }

                USBD_MSD_Task();
                if (_ReattchRequest) {
                        USB_SMSD_ReInit();
                        _ReattchRequest = 0;
                }
        }
}

/**
 * @brief Mass storage start
 */
void usb_smsd_start()
{
        OS_BASE_TYPE status;

        /* Start the USB SmartMSD application task. */
        status = OS_TASK_CREATE("UsbSmsdTask",  /* The text name assigned to the task, for
                                                   debug only; not used by the kernel. */
                        usb_smsd_task,          /* The function that implements the task. */
                        NULL,                   /* The parameter passed to the task. */
                        1024,                   /* The number of bytes to allocate to the
                                                     stack of the task. */
                        usb_main_TASK_PRIORITY, /* The priority assigned to the task. */
                        usb_smsd_task_handle);  /* The task handle. */

        OS_ASSERT(status == OS_TASK_CREATE_SUCCESS);
}

/**
 * @brief Mass storage stop
 */
void usb_smsd_stop()
{
        USBD_UnregisterSCHook(&UsbpHook);
        USBD_DeInit();
        OS_TASK_DELETE(usb_smsd_task_handle);
}

/**
 * @brief USB enumaration callback
 *    Event callback called from the usbcharger task to notify
 *    the application about to allow enumeration.
 *    Note: The USB charger task is started before the application task. Thus, these
 *          call-backs may be called before the application task is started.
 *          The application code should handle this case, if need be.
 */
void usb_start_enumeration_cb(void)
{
        if(run_usb_task == 0) {
                pm_stay_alive();
                _ReattchRequest = 0;
                run_usb_task = 1;
                isFwFile = false;
                cm_sys_clk_set(sysclk_PLL96);
                hw_usb_init();
                hw_usb_bus_attach();
                usb_smsd_start();
        }

}

/**
 * @brief USB detach callback
 *    Event callback called from the usbcharger task to notify
 *    the application that a detach of the USB cable was detected.
 *
 *    Note: The USB charger task is started before the application task. Thus, these
 *          call-backs may be called before the application task is started.
 *          The application code should handle this case, if need be.
 */
void usb_detach_cb(void)
{
       if(run_usb_task == 1) {
                hw_usb_bus_detach();
                usb_smsd_stop();
                cm_sys_clk_set(sysclk_XTAL16M);
                run_usb_task = 0;
                hw_usb_disable();
                pm_resume_sleep();
        }
}
